#include "subdivision.h"

Subdivision::Subdivision(QOpenGLFunctions_4_3_Core *f)
{
    this->f =f;
}

Subdivision::~Subdivision()
{
    delete shader;
}

void Subdivision::init() {
    shader = initComputeShaderProgram();
}

    QVector<unsigned int> Subdivision::fillVector(unsigned int a, unsigned int b, unsigned int c,unsigned int d){
        QVector<unsigned int> x;
        x.push_back(a);
        x.push_back(b);
        x.push_back(c);
        x.push_back(d);
        return x;
    }

QOpenGLShaderProgram *Subdivision::initComputeShaderProgram(){
    QString source = QLatin1String(":/subdivision.compute");

    qDebug()<<"Compiling compute shader ...";
    QOpenGLShader *computeShader = new QOpenGLShader(QOpenGLShader::Compute);
    if(!computeShader->compileSourceFile(source)){
        qCritical()<<"Compute Shader"<<source<<"failed"<<computeShader->log();
        exit(5);
    }

    qDebug()<<"Adding compute shader ...";
    QOpenGLShaderProgram *shader = new QOpenGLShaderProgram();
    shader->addShader(computeShader);
    qDebug()<<"Linking compute shader ...";
    if(!shader->link()){
        qCritical()<<"Linking compute shader failed:"<<shader->log();
        exit(5);
    }
    qDebug()<<"Linking compute shader done";

    return shader;
}

Mesh *Subdivision::subdivide(Mesh *mesh) {
    precomputeTables(mesh);
    runShader(mesh);
    return NULL;
}

void Subdivision::precomputeTables(Mesh *mesh) {
    Mesh::Node root = mesh->getRootNode();

    int firstMeshEntryIndex = -1;
    bool meshExists = root.getFirstMeshIndex(firstMeshEntryIndex);
    if (!meshExists){
        qDebug()<<"No Mesh found. Abort subdiv table precomputation.";
        return;
    }
    qDebug()<<"Mesh found. Index:"<<firstMeshEntryIndex;

    //die drei zeilen machen das bild schwarz. wenn man davor returnt tuts.
    Mesh::MeshEntry *firstMeshEntry = mesh->getMeshEntry(firstMeshEntryIndex);
    QVector<unsigned int> ib = firstMeshEntry->getIndexBuffer();
    qDebug()<<"Index Buffer: "<<ib;

    QVector<Vertex> vb = firstMeshEntry->getVertexBuffer();
    //qDebug()<<"Vertex Buffer: "<<vb;


    QVector<QVector<unsigned int>> edgeIndices_base;

    for (int i = 0; i < ib.length(); i+=3){
        //schaue alle dreiecke an

        Vertex x, y, z;
        unsigned int x_i = ib[i];
        unsigned int y_i = ib[i+1];
        unsigned int z_i = ib[i+2];
        x = vb[x_i];//todo maybe check array length
        y = vb[y_i];
        z = vb[z_i];

        //get vertices x,y,z from vertex buffer

        for (int j = i+3; j < ib.length(); j+= 3){
            //search all following triangles for common edge
            Vertex a,b,c;
            unsigned int a_i, b_i, c_i;
            a_i = ib[j];
            b_i = ib[j+1];
            c_i = ib[j+2];

            a = vb[a_i];
            b = vb[b_i];
            c = vb[c_i];

            //get vertices for a,b,c from vertex buffer
            //comparisons: xy = ba, cb, ac.
            //yz = ba, cb, ac
            //zx = ba, cb, ac
            //TODO if oder else if?
            if (x.samePos(a) && y.samePos(c)){
                edgeIndices_base.push_back(fillVector(x_i,y_i,z_i,b_i));
            }
            if (x.samePos(a) && z.samePos(b)){
                edgeIndices_base.push_back(fillVector(x_i,z_i,y_i,c_i));
            }
            if (x.samePos(b) && y.samePos(a)){
                edgeIndices_base.push_back(fillVector(x_i,y_i,z_i,c_i));
            }
            if (x.samePos(b) && z.samePos(c)){
                edgeIndices_base.push_back(fillVector(x_i,z_i,y_i,a_i));
            }
            if (x.samePos(c) && y.samePos(b)){
                edgeIndices_base.push_back(fillVector(x_i,y_i,z_i,a_i));
            }
            if (x.samePos(c) && z.samePos(a)){
                edgeIndices_base.push_back(fillVector(x_i,z_i,y_i,b_i));
            }

            if (y.samePos(a) && z.samePos(c)){
                edgeIndices_base.push_back(fillVector(y_i,z_i,x_i,b_i));
            }
            if (y.samePos(b) && z.samePos(a)){
                edgeIndices_base.push_back(fillVector(y_i,z_i,x_i,c_i));
            }
            if (y.samePos(c) && z.samePos(b)){
                edgeIndices_base.push_back(fillVector(y_i,z_i,x_i,a_i));
            }
        }//quadratische Laufzeit ftw
    }

    qDebug()<<"Done with edge table. "<<edgeIndices_base.length()<<" edges found. Table: "<<edgeIndices_base;
    if (true){
        qDebug()<<"Table (long version):";
        for (int i = 0; i < edgeIndices_base.length(); i++){
            qDebug()<<"blub:"<<edgeIndices_base[i];
        }
    }

    //TODO save/return edge table
}

void Subdivision::runShader(Mesh *mesh) {
    /*
     * This shader will set the output buffer to its global ID times hundred + the local id + offset.
     *
     * So, for offset = 0 we will get 0, 101, 200, 301 and so on as there's two threads in a local group.
     */

    qDebug()<<"Running compute shader";
    int num = 6 * 2; // must be a multiple of two as we do not handle the other case
    GLfloat offset = 3.0f;

    shader->bind();

    // Create an input buffer and set it to the value of offset
    GLuint offsetBufferID;
    f->glGenBuffers(1, &offsetBufferID);
    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 0, offsetBufferID);
    f->glBufferData(GL_SHADER_STORAGE_BUFFER, sizeof(GLfloat), &offset, GL_DYNAMIC_DRAW);

    // Create an output buffer
    GLuint bufferID;
    f->glGenBuffers(1, &bufferID);
    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 1, bufferID);
    f->glBufferData(GL_SHADER_STORAGE_BUFFER, sizeof(GLfloat) * num, NULL, GL_DYNAMIC_DRAW);

    // Run the shader
    f->glDispatchCompute(num/2, 1, 1);

    // Wait for the shader to complete and the data to be written back to the global memory
    f->glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);

    // Unbind the buffers from the shader
    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 0, 0);
    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 1, 0);

    // Map the output buffer so we can read the results
    f->glBindBuffer(GL_SHADER_STORAGE_BUFFER, bufferID);
    GLfloat *ptr;
    ptr = (GLfloat *) f->glMapBuffer(GL_SHADER_STORAGE_BUFFER, GL_WRITE_ONLY);
    for (int i = 0; i < num; i++) {
        qDebug()<<i<<ptr[i];
    }
    f->glUnmapBuffer(GL_SHADER_STORAGE_BUFFER);

    // Delete the buffers the free the resources
    f->glDeleteBuffers(1, &offsetBufferID);
    f->glDeleteBuffers(1, &bufferID);

    shader->release();
}
